//---------------------------------------------------------------------------
// Organization:    Graphical Communication, IFP, DTU
// Context:         Exercises, 67277, Computer Graphics
//
// Created/revised: July 1998
// File:            vector.h
// Purpose:
//
//
//---------------------------------------------------------------------------


#ifndef __VECTOR_H
#define __VECTOR_H

#include <stdlib.h>
#include <math.h>
#include "default.h"

//---------------------------------------------------------------------------
// class Vector
//
// This class represents a homogenouns 3D vector.
//---------------------------------------------------------------------------

class Vector 
{
  double a,b,c,d;

  friend class ostream& operator<<( class ostream& os, const Vector& x );
  friend inline double dot( const Vector &x, const Vector &y );
  friend inline double dot( const Vector &x, const double v[3] );
  friend inline Vector cross( const Vector &x, const Vector &y );
  friend class Matrix;

public:
  
  Vector( const double _a, const double _b, const double _c );
  Vector( const double ar[3]);
    Vector( const double _a, 
	    const double _b, 
	    const double _c, 
	    const double _d );
  Vector(void);
  
  Vector operator + ( const Vector &x ) const;
  Vector operator - ( const Vector &x ) const;
  Vector operator - ( void ) const;
  Vector operator * ( const double &k ) const;
  
  void operator -= ( const Vector &x );
  void operator += ( const Vector &x );
  void operator *= ( const Vector &x );
  void operator *= ( const double &k );
  void operator /= ( const double &k );
  
  int operator < ( const Vector &x ) const; 
  int operator <= ( const Vector &x ) const;
  int operator > ( const Vector &x ) const;
  int operator >= ( const Vector &x ) const;
  int operator != ( const int &val ) const;
  int operator != ( const Vector &v ) const;
  int operator == ( const Vector &v ) const;

  
  void cross( const Vector &x, const Vector &y );
  double len( void ) const;
  void normalize( void );
  void homogenize()
    {
      a = a/d;
      b = b/d;
      c = c/d;
      d = 1;
    };
  
  void invert();

  void get_spherical( double &theta, double &phi, double &rlen ) const;
  int set_spherical( const double &theta, 
		     const double &phi, 
		     const double &rlen );

  void copy_to_array( double dest[3] ) const;
  void get_ortogonal( Vector &x ) const;
  double& operator [] ( const int index );
  double operator [] ( const int index ) const;
  
};

//---------------------------------------------------------------------------
// The NULLVECTOR constant.
//---------------------------------------------------------------------------

const Vector NULLVECTOR;

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector::Vector(void) 
{
  a=b=c=0.0; 
  d=1.0;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
 
inline Vector::Vector( const double _a, const double _b, const double _c ):
  a(_a),b(_b),c(_c)
{
  d = 1.0;
}


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector::Vector( const double ar[3]) : a(ar[0]), b(ar[1]), c(ar[2]) 
{
  d = 1.0;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
 
inline Vector::Vector( const double _a, const double _b, 
		       const double _c, const double _d):
  a(_a),b(_b),c(_c),d(_d) {}


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector Vector::operator + ( const Vector &x ) const 
{
  return Vector( x.a+a, x.b+b, x.c+c);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector Vector::operator - ( const Vector &x ) const 
{
  return Vector( a-x.a, b-x.b, c-x.c );
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector Vector::operator - ( void ) const 
{
  return Vector( -a, -b, -c ); 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector Vector::operator * ( const double &k ) const 
{
  return Vector( a*k, b*k, c*k );
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::operator -= ( const Vector &x ) 
{ 
  a-=x.a; 
  b-=x.b; 
  c-=x.c; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::operator += ( const Vector &x ) 
{
  a+=x.a; 
  b+=x.b; 
  c+=x.c; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::operator *= ( const Vector &x ) 
{ 
  a*=x.a; 
  b*=x.b; 
  c*=x.c; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::operator *= ( const double &k ) 
{ 
  a*=k; 
  b*=k; 
  c*=k; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::operator /= ( const double &k ) 
{ 
  a/=k; 
  b/=k; 
  c/=k; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator < ( const Vector &x ) const 
{
  return (a<x.a && b<x.b && c<x.c);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator <= ( const Vector &x ) const 
{
  return (a<=x.a && b<=x.b && c<=x.c);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator > ( const Vector &x ) const 
{
  return (a>x.a && b>x.b && c>x.c);
}


//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator >= ( const Vector &x ) const 
{
  return (a>=x.a && b>=x.b && c>=x.c);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator != ( const int &val ) const 
{
  if (val==0 && a==0 && b==0 && c==0)
    return 0;
  else
    return 1;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator != ( const Vector &v ) const 
{
  if (a!=v.a || b!=v.b || c!= v.c)
    return 1;
  else
    return 0;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline int Vector::operator == ( const Vector &v ) const 
{
  if (a==v.a && b==v.b && c== v.c)
    return 1;
  else
    return 0;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::cross( const Vector &x, const Vector &y ) 
{
  a = x.b*y.c-x.c*y.b;
  b = x.c*y.a-x.a*y.c;
  c = x.a*y.b-x.b*y.a;
  d = 1;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
	

inline double Vector::len( void ) const 
{ 
  return sqrt( a*a + b*b + c*c ); 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::normalize( void ) 
{
  double len;
  if ((len = sqrt(a*a+b*b+c*c)) == 0.0)
    {
      cerr << "Can't normalize nullvector!" << endl;
      exit(1);
    }
  double invlen = 1.0/len;
  a *= invlen; b *= invlen; c *= invlen;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------


inline void Vector::invert() 
{ 
  a = -a; 
  b = -b; 
  c = -c; 
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline void Vector::copy_to_array( double dest[3] ) const 
{
  dest[0] = a; dest[1] = b; dest[2] = c;
}


//---------------------------------------------------------------------------
//  [] operator
//
// There are two identical versions of this function!
// The one above allows you to assign a value to a coordinate
// and the one below doesn't since it is a const function.
//
//---------------------------------------------------------------------------

inline double& Vector::operator [] ( const int index )
{
  switch( index ) {
  case 0: return a;
  case 1: return b;
  case 2: return c;
  case 3: return d;
  default:
    cerr <<  "Vector index must be in [0,3]";
    exit(1);
  }
  return a; // To avoid warnings, we don't get here
}


inline double Vector::operator [] ( const int index ) const
{
  switch( index ) 
    {
    case 0: return a;
    case 1: return b;
    case 2: return c;
    case 3: return d;
    default:
      cerr << "Vector index must be in [0,3]";
      exit(1);
    }
  return 0.0;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline double dot( const Vector &x, const Vector &y ) 
{
  return x.a*y.a + x.b*y.b + x.c*y.c;
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline double dot( const Vector &x, const double v[3] ) 
{
  return x.a*v[0] + x.b*v[1] + x.c*v[2];
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

inline Vector cross( const Vector &x, const Vector &y ) 
{
  return Vector( x.b*y.c-x.c*y.b, x.c*y.a-x.a*y.c, x.a*y.b-x.b*y.a );
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

ostream& operator<<( class ostream& os, const Vector& x );



#endif



